import Clutter from 'gi://Clutter';
import GObject from 'gi://GObject';
import Gio from 'gi://Gio';
import Meta from 'gi://Meta';
import St from 'gi://St';
import { Extension } from 'resource:///org/gnome/shell/extensions/extension.js';
import * as Main from 'resource:///org/gnome/shell/ui/main.js';
import Connections from './lib/connections.js';
import { Keys } from './lib/keys.js';
const EFFECT_NAME = 'window-desaturation-effect';
const TRANSITION_NAME = `${EFFECT_NAME}-factor`;
function gvalue(value) {
    const gv = new GObject.Value();
    gv.init(GObject.TYPE_DOUBLE);
    gv.set_double(value);
    return gv;
}
export default class WindowDesaturation extends Extension {
    connections;
    focused_window;
    pointed_window;
    settings;
    enable() {
        this.settings = this.getSettings();
        this.focused_window = this.window_with_focus();
        this.connections = new Connections();
        for (const window of this.windows()) {
            this.handle(window);
        }
        this.settings.connect('changed', () => this.refresh());
        const cursor = Meta.CursorTracker.get_for_display(global.display);
        this.connections.connect(cursor)('position-invalidated', () => this.update_pointer());
        this.connections.connect(global.display)('window-created', (_, window) => this.handle(window));
        this.connections.connect(global.display)('focus-window', (_, window) => this.update_focus(window));
        this.connections.connect(Main.overview)('showing', () => this.refresh());
        this.connections.connect(Main.overview)('hidden', () => this.refresh());
    }
    disable() {
        this.connections?.disconnect_all();
        this.connections = undefined;
        this.settings = undefined;
        for (const window of this.windows()) {
            this.clear(window);
        }
    }
    handle(window) {
        if (window.window_type !== Meta.WindowType.NORMAL)
            return;
        const actor = this.actor(window);
        if (!actor)
            return;
        const factor = this.factor(window);
        const effect = new Clutter.DesaturateEffect({ factor });
        actor.add_effect_with_name(EFFECT_NAME, effect);
    }
    clear(window) {
        const actor = this.actor(window);
        if (!actor)
            return;
        actor.remove_effect_by_name(EFFECT_NAME);
    }
    refresh() {
        this.focused_window = this.window_with_focus();
        this.pointed_window = this.window_with_pointer();
        for (const window of this.windows()) {
            this.update(window);
        }
    }
    update_focus(window) {
        if (this.focused_window === window)
            return;
        const before = this.focused_window;
        this.focused_window = window;
        this.update(before);
        this.update(window);
    }
    update_pointer() {
        const window = this.window_with_pointer();
        if (this.pointed_window === window)
            return;
        const before = this.pointed_window;
        this.pointed_window = window;
        this.update(before);
        this.update(window);
    }
    update(window) {
        const effect = this.effect(window);
        if (!effect)
            return;
        const actor = effect.actor;
        const enabled = this.should_affect(window);
        const factor = this.factor(window);
        actor.remove_transition(TRANSITION_NAME);
        effect.enabled = enabled;
        if (!enabled || !this.enable_animations) {
            effect.factor = factor;
            return;
        }
        const transition = new Clutter.PropertyTransition({
            actor: actor,
            property_name: `@effects.${effect.name}.factor`,
            remove_on_complete: true,
            duration: 250 * this.slow_down_factor,
            progress_mode: Clutter.AnimationMode.EASE_OUT_CUBIC,
        });
        transition.set_from(gvalue(effect.factor));
        transition.set_to(gvalue(factor));
        actor.add_transition(TRANSITION_NAME, transition);
    }
    actor(window) {
        if (!window)
            return;
        return window.get_compositor_private();
    }
    effect(window) {
        if (!window)
            return;
        const actor = this.actor(window);
        if (!actor)
            return;
        const effect = actor.get_effect(EFFECT_NAME);
        if (!(effect instanceof Clutter.DesaturateEffect))
            return;
        return effect;
    }
    should_affect(window) {
        if (Main.overview.visible && !this.handle_overview)
            return false;
        if (!window)
            return false;
        return !this.ignored_applications.includes(window.wm_class);
    }
    factor(window) {
        if (!window)
            return 0;
        if (window === this.focused_window)
            return this.foreground_factor;
        if (window === this.pointed_window && this.handle_hover)
            return this.hover_factor;
        return this.background_factor;
    }
    windows() {
        return global.display.list_all_windows();
    }
    window_with_focus() {
        return global.display.focus_window;
    }
    window_with_pointer() {
        return this.windows().find((it) => it.has_pointer());
    }
    get background_factor() {
        // biome-ignore lint/style/noNonNullAssertion: this.settings is defined
        return this.settings.get_double(Keys.background_factor);
    }
    get foreground_factor() {
        // biome-ignore lint/style/noNonNullAssertion: this.settings is defined
        return this.settings.get_double(Keys.foreground_factor);
    }
    get hover_factor() {
        // biome-ignore lint/style/noNonNullAssertion: this.settings is defined
        return this.settings.get_double(Keys.hover_factor);
    }
    get handle_overview() {
        // biome-ignore lint/style/noNonNullAssertion: this.settings is defined
        return this.settings.get_boolean(Keys.desaturate_overview);
    }
    get handle_hover() {
        // biome-ignore lint/style/noNonNullAssertion: this.settings is defined
        return this.settings.get_boolean(Keys.desaturate_hover);
    }
    get ignored_applications() {
        // biome-ignore lint/style/noNonNullAssertion: this.settings is defined
        return this.settings.get_strv(Keys.ignored_applications);
    }
    get enable_animations() {
        const settings = Gio.Settings.new('org.gnome.desktop.interface');
        return settings.get_boolean('enable-animations');
    }
    get slow_down_factor() {
        return St.Settings.get().slow_down_factor;
    }
}
